
/*-----------------------------------------------------------------------*
-- EXPANDER.C
-- 
-- Expand keywords into longer texts, while keeping indentation
-- 
-- Author: Marc verdiesen
-- Date:   15-Feb-96
--
-------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>


#define ISSPACE(a) ((a)==' ' || (a)=='\t')

void     chop_keyword( const char *original_line,
                       char       *truncated_line,
                       char       *keyword_string,
                       int        *indentation
                     );                 
int      check_keyword( const char *keyword_string
                       ,int        *table_index
                      );
                    
/*-----------------------------------------------------------------------*
-- KEYWORD EXPANSION FUNCTIONS
-- 
-- Each function accepts a single parameter, the indentation level to use.
-- Rules/assumptions for each of the expansion functions are :
--   - all text before the detected keyword has already been written
--   - the keyword has not been witten yet
--   - each expansion function will write at least one placeholder (@@@)
-------------------------------------------------------------------------*/
void expand_if       (int indent);
void expand_else     (int indent);
void expand_for      (int indent);
void expand_while    (int indent);
void expand_switch   (int indent);
void expand_case     (int indent);
void expand_default  (int indent);
void expand_enum     (int indent);
void expand_struct   (int indent);
void expand_funchead (int indent);
void expand_function (int indent);
void expand_block    (int indent);
void expand_include  (int indent);
void expand_cb       (int indent);


/*-----------------------------------------------------------------------*
-- KEYWORD TABLE
-------------------------------------------------------------------------*/
struct
{
   char     *key_name;
   void     (*expand_func)(int);
} 
keyword_table[] =
{ { "if"        ,expand_if       }     
 ,{ "else"      ,expand_else     }  
 ,{ "for"       ,expand_for      }
 ,{ "while"     ,expand_while    }
 ,{ "switch"    ,expand_switch   }
 ,{ "case"      ,expand_case     }
 ,{ "default"   ,expand_default  }
 ,{ "enum"      ,expand_enum     }     /* typedef enum             */
 ,{ "struct"    ,expand_struct   }     /* typedef struct           */
 ,{ "head"      ,expand_funchead }     /* function header          */
 ,{ "function"  ,expand_function }     /* function template        */
 ,{ "{"         ,expand_block    }     /* block                    */
 ,{ "#include"  ,expand_include  }     /* #include                 */
 ,{ "cb"        ,expand_cb       }     /* multi-line comment block */
 ,{ NULL        ,NULL            }     /* MUST BE THE LAST!!       */
};

                   

/*#------------------------------------------------------------------#*/
int main( void )
/*
 * Inputs   : none
 *
 * Outputs  : none
 *
 * InOuts   : none
 *
 * Returns  : process status (always 0)
 *
 * Purpose  : 'main' function of the program
 *
 * Notes    : Preconditions:
 *               there is a single line on stdin
 *            Postconditions:
 *               the input line has been processed to one or more output
 *               lines, according to the keyword found on the input line.
 */
{
   char     original_line[200];  
   char     truncated_line[200];
   char     keyword_string[200];
   int      indentation;
   int      table_index;
   

   /*--------------------------------------------------------------------*     
   -- Read the input line from stdin
   ----------------------------------------------------------------------*/
   gets(original_line);
   
   /*--------------------------------------------------------------------*
   -- Remove the keyword from the end of the original line and save the
   -- truncated remaining line.
   -- Determine the indentation of the keyword.
   ----------------------------------------------------------------------*/
   chop_keyword(  original_line,
                  truncated_line,
                  keyword_string,
                 &indentation
               );
               
   /*--------------------------------------------------------------------*
   -- Restore the truncated line
   ----------------------------------------------------------------------*/
   printf( "%s", truncated_line);
   
   /*--------------------------------------------------------------------*
   -- Determine if the keyword was a valid keyword.
   -- If it is not, restore it and append a placeholder (for repositioning
   -- the cursor in the editor).
   -- If it is, expand the keyword by calling its associated expansion
   -- function
   ----------------------------------------------------------------------*/
   if (check_keyword(keyword_string, &table_index))
   {
      (*(keyword_table[table_index].expand_func))(indentation);
   }
   else
   {
      printf("%s@@@", keyword_string);
   }
   
   return 0;
}

/*#------------------------------------------------------------------#*/
void     chop_keyword( const char *original_line,
                       char       *truncated_line,
                       char       *keyword_string,
                       int        *indentation
                     )
/*
 * Inputs   : original line      original input line
 *
 * Outputs  : truncated_line     original input line without the keyword
 *            keyword_string     keyword taken from the end of the original
 *                               line
 *            indentation        indentation at which the keyword was found
 *
 * InOuts   : none
 *
 * Returns  : none
 *
 * Purpose  : Take a possible keyword (any word) from the original line
 *            and determine its indentation.
 *
 * Notes    : Preconditions:
 *               'truncated_line' and 'keyword_string' are of at least the same
 *               size as 'original_line'.
 *            Postconditions:
 *               
 */

{
   int org_len;
   int trunc_len;
   int key_len;
   int i;
   int indent;
   int index;

   /*--------------------------------------------------------------------*
   -- Start truncated line as a copy of original line
   ----------------------------------------------------------------------*/
   strcpy( truncated_line, original_line );

   /*--------------------------------------------------------------------*
   -- Find the keyword, walking back from the end of the string until
   -- white space is found or the start of the string is reached
   ----------------------------------------------------------------------*/
   org_len   = strlen(original_line);
   trunc_len = org_len;

   while (trunc_len > 0 && !ISSPACE(truncated_line[trunc_len-1])) 
   {
      trunc_len--;
   }
   
   /*--------------------------------------------------------------------*
   -- Remove the keyword from the truncated line
   ----------------------------------------------------------------------*/
   truncated_line[trunc_len] = '\0';

   /*--------------------------------------------------------------------*
   -- Copy the keyword from the original line
   ----------------------------------------------------------------------*/
   key_len = org_len - trunc_len;
   for (i=0; i<key_len; i++)
   {
      keyword_string[i] = original_line[trunc_len+i];
   }
   keyword_string[i] = '\0';

   /*----------------------------------------------------------------------
   -- Determine the indentation of the keyword
   -- (minding the tabs)
   ----------------------------------------------------------------------*/
   indent = 0;
   index  = 0;
   while (truncated_line[index] != '\0')
   {
      if (truncated_line[index] == '\t')
      {
         indent = ((indent/8)+1)*8;
      }
      else
      {
         indent = indent +1;
      }
      index++;
   }
   *indentation = indent;
      
}    

/*#-----------------------------------------------------------------#*/
int check_keyword( const char *keyword_string
                  ,int        *table_index
                 )
/*
 * Inputs   : keyword_string     
 *
 * Outputs  : table_index     
 *
 * InOuts   : none
 *
 * Returns  : 0      keyword not found in list of keywords
 *            <> 0   keyword found
 *
 * Purpose  : Check if the keyword_string is the starting string of one of
 *            the keywords in keyword_table.
 *
 * Notes    : Preconditions:
 *               none
 *            Postconditions:
 *               return value = 
 *                   0     keyword not found, 'table_index' is undefined.
 *                   <>0   keyword found, 'table_index' is the index of the
 *                         keyword in 'keyword_table'.
 */         
{
   char     *full_keyword;
   char     *found_string;
   int       i;
   int       found = 0;

   /*--------------------------------------------------------------------*
   -- Lookup the keyword in the keyword_table.
   -- An empty keyword is caught first, because strstr will always
   -- say it is present in the string to be searched in.
   ----------------------------------------------------------------------*/
   if (keyword_string[0] != '\0')
   {
      i = 0;
      while(!found && keyword_table[i].key_name != NULL)
      {
         full_keyword = keyword_table[i].key_name;
         found_string = strstr( full_keyword, keyword_string);
         if (found_string == full_keyword)
         {
            found        = 1;
            *table_index = i;
         }
         i++;
      }
   }
   
   return found;
}



void print_block( int indentation )
{
   printf( "%*s{\n%*s   @@@\n%*s}"
          ,indentation, ""
          ,indentation, ""
          ,indentation, ""
         );
}


/*-----------------------------------------------------------------------*
-- EXPANSION FUNCTIONS
-------------------------------------------------------------------------*/
void expand_if( int indentation )
{
   printf( "if (@@@)\n");
   print_block(indentation);   
}

void expand_else( int indentation )
{
   printf( "else\n");
   print_block(indentation);   
}

void expand_for( int indentation )
{
   printf( "for (@@@; ; )\n");
   print_block(indentation); 
}

void expand_while( int indentation )
{
   printf( "while (@@@)\n");
   print_block(indentation);   
}


void expand_switch( int indentation )
{
   printf( "switch (@@@)\n"
           "%*s{\n"
           "%*s   case @@@:\n"
           "%*s      @@@\n"
           "%*s   case @@@:\n"
           "%*s      @@@\n"
           "%*s   default:\n"
           "%*s      @@@\n"
           "%*s}\n"
          ,indentation, ""
          ,indentation, ""
          ,indentation, ""
          ,indentation, ""
          ,indentation, ""
          ,indentation, ""
          ,indentation, ""
          ,indentation, ""
         );
}


void expand_case( int indentation )
{
   printf( "case @@@:\n"
           "%*s   @@@\n"
          ,indentation, ""
         );
}

void expand_default( int indentation )
{
   printf( "default:\n"
           "%*s   @@@\n"
          ,indentation, ""
         );
}


void expand_cb( int indentation )            /* comment block  */
{
   int i;
   
   
   printf("/*");
   for (i=indentation+4; i < 75; i++)
   {
      putchar( '-' );
   }
   printf("*\n");
   printf( "%*s-- @@@\n"
           "%*s--"
          ,indentation, ""
          ,indentation, ""
         );
   for (i=indentation+4; i < 75; i++)
   {
      putchar( '-' );
   }
   printf("*/");
}
                   

void expand_funchead( int indentation )      /* Function header */
{
   if (indentation > 0)
   {
      printf("\n");
   } 
   printf( "/*\n"
           " * Inputs   : @@@\n"
           " *\n"
           " * Outputs  : @@@\n"
           " *\n"
           " * InOuts   : @@@\n"
           " *\n"
           " * Returns  : @@@\n"
           " *\n"
           " * Purpose  : @@@\n" 
           " *\n"
           " * Notes    : Preconditions:\n"
           " *               @@@\n"
           " *            Postconditions:\n"
           " *               @@@\n"
           " */\n"
         );
}
                   

void expand_function( int indentation )      /* function template */
{
   if (indentation > 0)
   {
      printf("\n");
   }
   printf( "\n"
           "/*#------------------------------------------------------------------#*/\n"
           "@@@\n"
         );
   expand_funchead( indentation );
   printf( "{\n"
           "   @@@\n"
           "   return @@@;\n"
           "}\n"
         );
}

void expand_enum( int indentation )
{
   if (indentation > 0)
   {
      printf("\n");
   }
   printf( "typedef enum\n" );
   print_block(0);
   printf(" @@@;\n");
}
  

void expand_struct( int indentation )
{
   if (indentation > 0)
   {
      printf("\n");
   }
   printf( "typedef struct\n" );
   print_block(0);
   printf(" @@@;\n");
}
  

void expand_block( int indentation )
{
   printf( "{\n" );
   printf( "%*s   @@@\n"
           "%*s}"
          ,indentation, ""
          ,indentation, ""
         );
}

void expand_include( int indentation )
{
   if (indentation > 0)
   {
      printf("\n");
   }
   printf( "#include <@@@.h>       /* @@@ */" );
}
